#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/6/20 19:10
@File    : convert_coco91to81.py
@Software: PyCharm
@Desc    : 
"""
import argparse
import os
import os.path as osp

import numpy as np
from PIL import Image
from alchemy_cat.contrib.coco import COCO
from alchemy_cat.data.plugins import arr2PIL
from tqdm import tqdm

parser = argparse.ArgumentParser()
parser.add_argument('-s', '--source', type=str)
parser.add_argument('-t', '--target', type=str)
parser.add_argument('--label_type', type=str, default='')
parser.add_argument('--split', type=str, default='train')
args = parser.parse_args()

dt = COCO('datasets', split=args.split, label_type=args.label_type, cls_labels_type=None)

os.makedirs(args.target, exist_ok=True)

for img_id in tqdm(dt.image_ids, dynamic_ncols=True, desc="处理", unit="张"):
    label = np.array(Image.open(osp.join(args.source, f'{img_id}.png')), dtype=np.uint8)
    converted_lb = COCO.coco91to81[label]
    arr2PIL(converted_lb).save(osp.join(args.target, f'{img_id}.png'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/6/22 22:29
@File    : cfg.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.py_tools import Cfg2Tune, Param2Tune, ParamLazy

cfg = config = Cfg2Tune()

cfg.rslt_dir = 'param_lazy_config'

# * Param on same Cfg2Tune
cfg.foo0.foo1.a = Param2Tune([0, 1, 2])

# * Param Lazy
cfg.foo0.lazy0 = ParamLazy(lambda c: c.foo0.foo1.a + 10)
cfg.lazy1 = ParamLazy(lambda c: c.foo0.foo1.a - c.foo0.a_max)

# * Static Value
cfg.foo0.a_max = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: data_auger.py
@time: 2020/1/12 1:48
@desc:
"""
from typing import Union, Optional
import numpy as np
from inspect import signature
import lmdb
import pickle
from functools import reduce
from torch.utils import data as torch_data

from alchemy_cat.dag import Graph
from alchemy_cat.data import Dataset
from alchemy_cat.py_tools.str_formatters import indent
from alchemy_cat.py_tools.type import is_int
from alchemy_cat.alg import accumulate

__all__ = ["RandMap", "MultiMap", "DataAuger"]


class RandMap(object):
    rand_seeds = None  # Can be overloaded. The static rand seeds where rand seed is selected.
    weight = None  # Can be overloaded. The static weight to choose the rand seeds

    def __init__(self):
        self._rand_seed = None
        self._is_rand_seed_set = False

        # Check whether the arguments of self.forward is the same to the self.generate_rand_seed
        def get_param_names(func):
            return list(signature(func).parameters.keys())

        if get_param_names(self.generate_rand_seed) != get_param_names(self.__call__) and \
                get_param_names(self.generate_rand_seed) != get_param_names(self.forward):
            raise RuntimeError(f"{type(self).__name__}'s self.generate_rand_seed's args is not "
                               f"equal to the args of self.__call__ or self.forward")

    def __len__(self):
        return len(self.rand_seeds)

    def generate_rand_seed(self, *fwd_args, **fwd_kwargs):
        """Generate rand seed which will be used in forward.

        Can be overloaded to dynamically generate the rand seed according to the forward() inputs.
        But the arguments of this function should be exactly the same to the forward() function.

        Returns: Generated rand seed
        """
        prob = None
        if self.weight is not None:
            prob = np.array(self.weight, dtype=np.float32) / np.sum(self.weight)
        rand_seed_index = np.random.choice(len(self), p=prob)
        return self.rand_seeds[rand_seed_index]

    @property
    def rand_seed(self):
        if self._rand_seed is None:
            raise AttributeError(f"Object of {type(self).__name__} still don't have rand_seed yet")
        else:
            return self._rand_seed

    @rand_seed.setter
    def rand_seed(self, value):
        if self.rand_seeds is not None and value not in self.rand_seeds:
            raise KeyError(f"rand_seed value f{value} to be set is not in self.rand_seeds={self.rand_seeds}")

        self._rand_seed = value
        self._is_rand_seed_set = True

    def forward(self, *fwd_args, **fwd_kwargs):
        """Function which you wish to call when the functor is called. Need to be overloaded
        It should be implemented according to rand_seed
        """
        raise NotImplementedError

    def __call__(self, *fwd_args, **fwd_kwargs):
        if not self._is_rand_seed_set:
            self.rand_seed = self.generate_rand_seed(*fwd_args, **fwd_kwargs)

        ret = self.forward(*fwd_args, **fwd_kwargs)
        self._is_rand_seed_set = False
        return ret

    def __repr__(self):
        return f"RandMap: {self.__class__.__name__}\n" \
                + f"    _rand_seed: {self._rand_seed}\n" \
                + f"    _is_rand_seed_set: {self._is_rand_seed_set}"


class MultiMap(object):
    output_num = None

    def __init__(self):
        self._output_index = 0

    def __len__(self):
        if not is_int(self.output_num):
            raise ValueError(f"{self.__class__.__name__}'s output_num {self.output_num} is not an int value")
        return int(self.output_num)

    @property
    def output_index(self):
        return self._output_index

    @output_index.setter
    def output_index(self, value):
        if not is_int(value):
            raise ValueError(f"output_index {value} should be int")

        if value < 0:
            value += len(self)

        if value < 0 or value >= len(self):
            raise ValueError(f"output_index value {value} to be set is out of range [{0}, {self.output_num})")

        self._output_index = value

    def forward(self, *fwd_args, **fwd_kwargs):
        """Function which you wish to call when the functor is called. Need to be overloaded.
        It should be implemented according to output_index
        """
        raise NotImplementedError

    def __call__(self, *fwd_args, **fwd_kwargs):
        return self.forward(*fwd_args, **fwd_kwargs)

    def __repr__(self):
        return f"MultiMap: {self.__class__.__name__}\n" \
                + f"    output_num: {self.output_num}\n" \
                + f"    _output_index: {self._output_index}"


class DataAuger(Dataset):

    def __init__(self, dataset: Union[Dataset, torch_data.Dataset], verbosity: int = 0, pool_size: int = 0,
                 slim: bool = False, rand_seed_log: str = None):
        """Given dataset and argument it with a calculate graph.

        Args:
            dataset: dataset to be augmented
            verbosity: Calculate graph's verbosity. 0: No log output; 1: Only give graph level log;
                >1: Give graph and node level log.
            pool_size: Calculate graph's pool size. 0 means don't use parallel.
            slim: If True, calculate graph use copy rather than deepcopy when setting value of Node's input.
            rand_seed_log: lmdb database where rand seeds saved
        """
        self._dataset = dataset

        self._rand_seed_log: Optional[str] = None
        self.lmdb_env: Optional[lmdb.Environment] = None
        self.rand_seed_log = rand_seed_log

        self.graph = Graph(verbosity=verbosity, pool_size=pool_size, slim=slim)
        self.build_graph()

        self.deep_prefix_id_ordered_nodes = self.graph.deep_prefix_id_ordered_nodes()

        multi_prefix_id_nodes = [(prefix_id_node) for prefix_id_node in self.deep_prefix_id_ordered_nodes
                                 if isinstance(prefix_id_node[1]._fct, MultiMap)]
        self.multi_nodes_prefix_id, self.multi_nodes = zip(*multi_prefix_id_nodes) if multi_prefix_id_nodes \
            else ([], [])

        rand_prefix_id_nodes = [(prefix_id_node) for prefix_id_node in self.deep_prefix_id_ordered_nodes
                                if isinstance(prefix_id_node[1]._fct, RandMap)]
        self.rand_nodes_prefix_id, self.rand_nodes = zip(*rand_prefix_id_nodes) if rand_prefix_id_nodes else \
            ([], [])

        self.multi_factors = self._get_multi_factors()
        self.divide_factors = self._get_divide_factors()

    def build_graph(self):
        """Build self.graph here. Need to be overloaded"""
        raise NotImplementedError

    def _get_multi_factors(self):
        multi_factors = [len(self._dataset)]

        for node in self.multi_nodes:
            multi_factors.append(len(node._fct))

        return multi_factors

    def _get_divide_factors(self):
        divide_factors = self.multi_factors[::-1][:-1]
        divide_factors = list(accumulate(divide_factors, lambda x, y: x * y))
        return divide_factors[::-1] + [1]

    @property
    def dataset(self):
        """Get auger's dataset"""
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        """Reset auger's dataset"""
        self._dataset = value
        self.multi_factors = self._get_multi_factors()

    @property
    def multi_factor(self):
        """Return the number of one example were auger to"""
        return reduce(lambda x, y: x * y, self.multi_factors[1:])

    def __len__(self):
        """Return the size of augmented dataset"""
        return reduce(lambda x, y: x * y, self.multi_factors)

    def load_indices(self, idx):
        """Set every multi node's idx and return dataset's idx with auger's idx"""
        dataset_idx = idx // self.divide_factors[0]
        idx %= self.divide_factors[0]

        for multi_node, divide_factor in zip(self.multi_nodes, self.divide_factors[1:]):
            multi_node._fct.output_index = idx // divide_factor
            idx %= divide_factor

        return dataset_idx

    def calculate_indices(self, idx):
        """Return dataset's idx and every multi node's idx with auger's idx"""
        dataset_idx = idx // self.divide_factors[0]
        idx %= self.divide_factors[0]

        node_indices = {}
        for prefix_id, divide_factor in zip(self.multi_nodes_prefix_id, self.divide_factors[1:]):
            node_indices[prefix_id] = idx // divide_factor
            idx %= divide_factor

        return dataset_idx, node_indices

    @property
    def rand_seeds(self):
        """Return rand seed of every rand node"""
        rand_seeds = {}

        for prefix_id, node in zip(self.rand_nodes_prefix_id ,self.rand_nodes):
            rand_seeds[prefix_id] = node._fct.rand_seed
        return rand_seeds

    def load_rand_seeds(self, rand_seeds):
        """Load rand seed from rand_seeds"""
        for prefix_id, node in zip(self.rand_nodes_prefix_id ,self.rand_nodes):
            node._fct.rand_seed = rand_seeds[prefix_id]

    @property
    def rand_seed_log(self):
        return self._rand_seed_log

    @rand_seed_log.setter
    def rand_seed_log(self, rand_seed_log):
        if (not isinstance(rand_seed_log, str)) and (rand_seed_log is not None):
            raise ValueError(f"rand_seed_log = {rand_seed_log} must be str or None")

        self._rand_seed_log = rand_seed_log

        if self.lmdb_env is not None:
            self.lmdb_env.close()
            self.lmdb_env = None

        if rand_seed_log is not None:
            self.lmdb_env = lmdb.open(rand_seed_log, meminit=False, map_size=2147483648, max_spare_txns=64,
                                      sync=False, metasync=False, lock=True)

    def __del__(self):
        self.rand_seed_log = None

    def get_item(self, idx):
        example = self._dataset[self.load_indices(idx)]

        if self.lmdb_env is not None:
            with self.lmdb_env.begin() as txn:
                rand_seeds = txn.get(str(idx).encode())
            if rand_seeds is not None:
                rand_seeds = pickle.loads(rand_seeds)
                self.load_rand_seeds(rand_seeds)

        ret = self.graph.calculate(data={'example': example})

        if self.lmdb_env is not None and rand_seeds is None:
            with self.lmdb_env.begin(write=True) as txn:
                txn.put(str(idx).encode(), pickle.dumps(self.rand_seeds))

        return ret

    def __repr__(self):
        return f"DataAuger <{self.__class__.__name__}>:\n" \
                + indent(f"graph: {self.graph}") + "\n" \
                + indent(f"dataset: {self.dataset}") + "\n" \
                + indent(f"#DataAuger: {len(self)}")



# -- configs/addons/linear_warm_cos_sched.py --
import torch.optim.lr_scheduler as sched

from alchemy_cat.dl_config import Config, DEP

cfg = Config()

cfg.epochs = 30

@cfg.set_DEP(priority=0)  # warm_epochs = 10% of total epochs
def warm_epochs(c: Config) -> int:
    return round(0.1 * c.epochs)

cfg.warm.cls = sched.LinearLR
cfg.warm.ini.total_iters = DEP(lambda c: c.warm_epochs, priority=1)
cfg.warm.ini.start_factor = 1e-5
cfg.warm.ini.end_factor = 1.

cfg.main.cls = sched.CosineAnnealingLR
cfg.main.ini.T_max = DEP(lambda c: c.epochs - c.warm.ini.total_iters,
                         priority=2)  # main_epochs = total_epochs - warm_epochs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: utils.py
@time: 2020/3/20 2:07
@desc:
"""
from typing import Optional
import os
import lmdb
import pickle

__all__ = ['read_rand_seeds', 'write_rand_seeds']


def read_rand_seeds(rand_seed_log: str, idx: int, create: bool=False) -> Optional[dict]:
    """Read rand seeds from idx of rand_seed_log's

    Args:
        rand_seed_log: rand seed log file
        idx: Index of sample
        create: If True, lmdb database with path rand_seed_log will be created. (Default: False)

    Returns:
        If rand seeds exit, return rand seeds. Else return None.
    """
    if not create and not os.path.isdir(rand_seed_log):
        return None

    env = lmdb.open(rand_seed_log, meminit=False, readonly=True)

    with env.begin() as txn:
        rand_seeds = txn.get(str(idx).encode(), default=None)

    if rand_seeds is not None:
        rand_seeds = pickle.loads(rand_seeds)

    env.close()
    return rand_seeds


def write_rand_seeds(rand_seed_log: str, idx: int, rand_seeds: dict):
    """Write rand seeds to idx of rand_seed_log's

    Args:
        rand_seed_log: rand seed log file
        idx: Index of sample
        rand_seeds: rand seeds to be writen
    """
    env = lmdb.open(rand_seed_log, meminit=False)

    with env.begin(write=True) as txn:
        txn.put(str(idx).encode(), pickle.dumps(rand_seeds))

    env.close()



#!/usr/bin/env src
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: vgg_feature.py
@time: 2019/12/12 12:31
@desc: VGG Model. Based on torch VISION
"""
from typing import Dict

import torch
import torch.nn as nn

from alchemy_cat.alg import odd_input_pad_size

from alchemy_cat.torch_tools import update_model_state_dict, strip_named_params, strip_param_name
from alchemy_cat.torch_tools import Network

__all__ = ['ConvReLU', 'VGGLayer', 'VGGFeature']


class ConvReLU(nn.Sequential, Network):

    def __init__(self, in_ch, out_ch, dilation, layer_idx, seq_idx):
        super(ConvReLU, self).__init__()

        self.in_channel = in_ch
        self.out_channel = out_ch

        self.add_module(f"conv{layer_idx}_{seq_idx}",
                        nn.Conv2d(in_channels=in_ch, out_channels=out_ch, kernel_size=3, padding=dilation,
                                  dilation=dilation))
        self.add_module(f"relu{layer_idx}_{seq_idx}",
                        nn.ReLU(inplace=True))


class VGGLayer(nn.Sequential, Network):

    def __init__(self, in_ch, out_ch, conv_num, dilation, pool_size, pool_stride, layer_idx):
        super(VGGLayer, self).__init__()

        self.in_channel = in_ch
        self.out_channel = out_ch

        for seq_idx in range(1, conv_num+1):
            self.add_module(f"conv_relu_{seq_idx}",
                            ConvReLU(in_ch=in_ch if seq_idx == 1 else out_ch, out_ch=out_ch,
                                     dilation=dilation, layer_idx=layer_idx, seq_idx=seq_idx))

        padding = odd_input_pad_size(pool_size)

        self.add_module(f"pool{layer_idx}",
                        nn.MaxPool2d(kernel_size=pool_size, stride=pool_stride,
                                     padding=padding, ceil_mode=True))


class VGGFeature(nn.Sequential, Network):

    def __init__(self, in_ch, out_ch, conv_nums, dilations, pool_strides, pool_size):
        super(VGGFeature, self).__init__()

        self.in_channel = in_ch
        self.out_channel = out_ch[-1]

        for i, layer_idx in enumerate(range(1, len(out_ch) + 1)):
            self.add_module(f"layer{layer_idx}",
                            VGGLayer(in_ch=in_ch if layer_idx == 1 else out_ch[i - 1],
                                     out_ch=out_ch[i], conv_num=conv_nums[i], dilation=dilations[i],
                                     pool_size=pool_size[i], pool_stride=pool_strides[i], layer_idx=layer_idx))

    def initialize(self, pretrain_dict: Dict[str, nn.Parameter], verbosity: int=1, **kwargs):
        strip_pretrain_dict = strip_named_params(pretrain_dict)
        weight_dict = {}

        for model_param_name in self.state_dict().keys():
            strip_model_param_name = strip_param_name(model_param_name)
            if strip_model_param_name in strip_pretrain_dict:
                weight_dict[model_param_name] = strip_pretrain_dict[strip_model_param_name]

        update_model_state_dict(self, weight_dict, verbosity)


if __name__ == "__main__":
    from torch.utils.tensorboard import SummaryWriter
    writer=SummaryWriter(log_dir="Temp")

    writer.add_graph(VGGFeature(
        in_ch=3,
        out_ch=[64, 128, 256, 512, 512],
        conv_nums=[2, 2, 3, 3, 3],
        dilations=[1, 1, 1, 1, 2],
        pool_strides=[2, 2, 2, 1, 1],
        pool_size=[3, 3, 3, 1, 1]
    ), input_to_model=torch.randn((10, 3, 321, 321), dtype=torch.float32))

    print("Graph has been writen to the temp dir")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/13 16:20
@File    : eval_cams.py
@Software: PyCharm
@Desc    : 
"""
import glob
import multiprocessing as mp
import os
import pickle
import warnings
from collections import defaultdict
from functools import partial
from os import path as osp
from typing import Optional, Iterable, Callable, Tuple

import cv2
import numpy as np
from PIL import Image
from alchemy_cat.contrib.metrics import SegmentationMetric
from alchemy_cat.acplot import RGB2BGR
from alchemy_cat.data.plugins import identical
from alchemy_cat.py_tools import OneOffTracker
from alchemy_cat.dl_config import Config
from frozendict import frozendict as fd
from tqdm import tqdm

__all__ = ['eval_cams', 'search_and_eval']


def _eval_cams_core(cam_file_suffix: str,
                    cam2pred: Callable,
                    cam_dir: str, gt_dir: str, image_dir: str,
                    gt_preprocess: Callable[[np.ndarray], np.ndarray],
                    crf: Callable[[np.ndarray, np.ndarray], np.ndarray]
                    ) -> Tuple[str, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    img_id = osp.splitext(cam_file_suffix)[0]
    cam_file = osp.join(cam_dir, cam_file_suffix)
    gt_file = osp.join(gt_dir, f'{img_id}.png')

    # Read files
    gt = cv2.imread(gt_file, cv2.IMREAD_GRAYSCALE)
    assert gt is not None
    gt: np.ndarray = gt_preprocess(gt)

    loaded = np.load(cam_file)
    cam, fg_cls = loaded['cam'].astype(np.float32), loaded['fg_cls'].astype(np.uint8)

    if crf is not None:
        dsize = np.asarray(Image.open(osp.join(image_dir, f'{img_id}.jpg')), dtype=np.uint8)
    else:
        dsize = gt.shape

    pred, fg_bg_score = cam2pred(cam, dsize, fg_cls, crf=crf)

    assert pred.shape == gt.shape

    return img_id, pred, fg_bg_score, fg_cls, gt


def eval_cams(class_num: int, class_names: Optional[Iterable[str]],
              cam_dir: str, preds_ignore_label: int,
              gt_dir: str, gts_ignore_label: int,
              cam2pred: Callable,
              image_dir: str | None=None,
              crf: Callable[[np.ndarray, np.ndarray], np.ndarray] | None=None,
              result_dir: Optional[str]=None,
              gt_preprocess: Callable[[np.ndarray], np.ndarray]=identical,
              importance: int=2,
              eval_individually: bool=True, take_pred_ignore_as_a_cls: bool=False,
              metric_cls: type=SegmentationMetric,
              pool_size=0,
              save_sample_data: bool=False) \
        -> (SegmentationMetric, defaultdict):

    """Evaluate predictions of semantic segmentation

    Args:
        class_num: Num of classes
        class_names: Name of classes
        cam_dir: Dictionary where cam stored
        preds_ignore_label: Ignore label for predictions
        gt_dir: Dictionary where ground truths stored
        image_dir: Dictionary where images stored. (Default: None)
        gts_ignore_label: Ignore label for ground truths
        result_dir: If not None, eval result will be saved to result_dir. (Default: None)
        cam2pred: Function transfer cam file to prediction.
        crf: Function to refine prediction. (Default: None)
        gt_preprocess: Preprocess function for ground truth read in
        importance: Segmentation Metric's importance filter. (Default: 2)
        eval_individually: If True, evaluate each sample. (Default: True)
        take_pred_ignore_as_a_cls: If True, the ignored label in preds will be seemed as a class. (Default: False)
        metric_cls: Use metric_cls(class_num, class_names) to eval preds. (Default: SegmentationMetric)
        pool_size: Pool size for multiprocessing. (Default: 0)
        save_sample_data: If True, sample data (pred, bg_fg_score) will be saved. (Default: False)

    Returns:
        Segmentation Metric and metric result for each sample (If eval_individually is True)
    """
    assert preds_ignore_label >= class_num
    assert gts_ignore_label >= class_num

    if take_pred_ignore_as_a_cls:
        class_num += 1
        class_names += ['pred_ignore']

    metric = metric_cls(class_num, class_names)

    sample_data = defaultdict(dict)

    # * 构造核心函数。
    core = partial(_eval_cams_core,
                   cam2pred=cam2pred,
                   cam_dir=cam_dir, gt_dir=gt_dir, image_dir=image_dir,
                   gt_preprocess=gt_preprocess,
                   crf=crf)

    print("\n================================== Eval ==================================")
    cam_file_suffixes = os.listdir(cam_dir)

    if pool_size > 0:
        p = mp.Pool(pool_size)
        it = p.imap_unordered(core, cam_file_suffixes, chunksize=10)
    else:
        it = map(core, cam_file_suffixes)

    for img_id, pred, bg_fg_score, fg_cls, gt in tqdm(it,
                                                      total=len(cam_file_suffixes), miniters=10,
                                                      desc='eval progress', unit='sample', dynamic_ncols=True):

        if take_pred_ignore_as_a_cls:
            pred[pred == preds_ignore_label] = class_num - 1
            if gts_ignore_label == class_num - 1:
                gt[gt == gts_ignore_label] = class_num

        # Evaluate
        metric.update(pred, gt)
        if eval_individually:
            with OneOffTracker(lambda: metric_cls(class_num, class_names)) as individual_metric:
                individual_metric.update(pred, gt)
            sample_data[img_id]['metric'] = individual_metric.statistics(importance)

        if save_sample_data:
            sample_data[img_id]['seed'] = pred  # 在外面叫seed。
            sample_data[img_id]['bg_fg_score'] = bg_fg_score
            sample_data[img_id]['fg_cls'] = fg_cls

    if pool_size > 0:
        p.close()
        p.join()

    # Saving
    if result_dir is not None:
        metric.save_metric(result_dir, importance, dpi=400)
        if eval_individually:
            with open(osp.join(result_dir, 'sample_statistics.pkl'), 'wb') as f:
                pickle.dump({i: d['metric'] for i, d in sample_data.items()}, f)

    print("\n================================ Eval End ================================")

    return metric, sample_data


def _save_sample_data(sample_data: defaultdict | str, save_dir: str, mask_cfg: Config):
    data_dir = osp.join(save_dir, 'data')
    mask_dir = osp.join(save_dir, 'mask')
    viz_mask_dir = osp.join(save_dir, 'viz_mask')

    mask_only = False
    if sample_data == 'saved':
        assert osp.isdir(data_dir)  # 检查是否真的保存了。
        if mask_cfg and (not osp.isdir(mask_dir)):  # 虽然存过了，但mask没存。
            mask_only = True
            data_files = glob.glob(osp.join(data_dir, '*.npz'))
            sample_data = {osp.splitext(osp.basename(f))[0]: np.load(f) for f in data_files}
        else:
            return

    os.makedirs(data_dir, exist_ok=True)
    if mask_cfg:
        os.makedirs(mask_dir, exist_ok=True)
        os.makedirs(viz_mask_dir, exist_ok=True)

    for img_id, data in sample_data.items():
        if not mask_only:
            np.savez(osp.join(data_dir, f'{img_id}.npz'),
                     seed=data['seed'], bg_fg_score=data['bg_fg_score'], fg_cls=data['fg_cls'])
        if mask_cfg:
            cv2.imwrite(osp.join(mask_dir, f'{img_id}.png'), mask := mask_cfg.cal(data).astype(np.uint8))
            cv2.imwrite(osp.join(viz_mask_dir, f'{img_id}.png'), RGB2BGR(mask_cfg.viz(mask)))


def search_and_eval(dt, cam_dir: str, seed_cfg: Config, rslt_dir: str, pool_size: int=0):
    """搜索不同bg_method，获取最优性能。

    Args:
        dt: 数据集，提供类别数、类别名、忽略标签、标签目录等信息。
        cam_dir: cam文件所在目录。
        rslt_dir: 结果的目录。
        seed_cfg: seed_cfg.cal(cam, gt.shape, fg_cls, seed_cfg.bg_methods[i], **seed_cfg.ini)将CAM转换为seed。
        pool_size: 并行计算的进程数。0表示不使用并行计算。 (Default: 0)

    Returns:
        None
    """
    os.makedirs(eval_dir := osp.join(rslt_dir, 'eval'), exist_ok=True)
    if seed_cfg.save is not None:
        assert seed_cfg.save in ['all', 'best']
        os.makedirs(seed_dir := osp.join(rslt_dir, 'seed'), exist_ok=True)

    # * 若已经有bg_method_metrics.pkl，则直接读取。
    if osp.isfile(bg_method_metrics_pkl := osp.join(eval_dir, 'bg_method_metrics.pkl')):
        with open(bg_method_metrics_pkl, 'rb') as f:
            bg_method_metrics = pickle.load(f)
        if seed_cfg.save is not None:
            warnings.warn(f"应当确保缓存的{bg_method_metrics_pkl}保存了对应的种子。")
    else:
        bg_method_metrics = {}

    # * 对各配置中的methods，计算其metric。
    best_bg_method, best_sample_data = None, None

    for bg_method in seed_cfg.bg_methods:
        bg_method = fd(dict(bg_method))  # 将dict转换为frozendict，以便作为字典的key。

        if bg_method in bg_method_metrics:
            metric, sample_data = bg_method_metrics[bg_method], 'saved'
        else:
            metric, sample_data = eval_cams(class_num=dt.class_num,
                                            class_names=dt.class_names,
                                            cam_dir=cam_dir,
                                            preds_ignore_label=255,
                                            gt_dir=dt.label_dir,
                                            gts_ignore_label=dt.ignore_label,
                                            cam2pred=partial(seed_cfg.cal,
                                                             bg_method=bg_method, **seed_cfg.ini),
                                            image_dir=dt.image_dir,
                                            crf=seed_cfg.crf,
                                            result_dir=None,
                                            importance=0,
                                            eval_individually=False,
                                            take_pred_ignore_as_a_cls=False,
                                            pool_size=pool_size,
                                            save_sample_data=seed_cfg.save is not None)

        print(f'Current mIoU: {metric.mIoU:.4f} (bg_method={bg_method})')
        bg_method_metrics[bg_method] = metric

        if (best_bg_method is None) or (metric.mIoU > bg_method_metrics[best_bg_method].mIoU):
            best_bg_method, best_sample_data = bg_method, sample_data

        if seed_cfg.save == 'all':
            _save_sample_data(sample_data, osp.join(seed_dir, f'{metric.mIoU:.2f}'), seed_cfg.mask)

    # * 保存method_metrics.pkl。
    with open(bg_method_metrics_pkl, 'wb') as f:
        pickle.dump(bg_method_metrics, f)

    # * 打印所有评价结果。
    for bg_method, metric in bg_method_metrics.items():
        print(f'mIoU: {metric.mIoU:.4f} (bg_method={bg_method})')
    print(f'Best mIoU: {bg_method_metrics[best_bg_method].mIoU:.4f} (bg_method={best_bg_method})')

    # * 遍历method_metrics字典，找到最好的metric。
    bg_method_metrics[best_bg_method].save_metric(eval_dir, importance=0, figsize=(24, 24))
    if seed_cfg.save == 'best':
        _save_sample_data(best_sample_data, osp.join(seed_dir, 'best'), seed_cfg.mask)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/4/17 14:47
@File    : resolve_loc_cue_conflict.py
@Software: PyCharm
@Desc    : 
"""
from typing import List
import numpy as np

from alchemy_cat.py_tools.type import is_intarr

__all__ = ['resolve_loc_cue_conflict_by_priority', 'resolve_loc_cue_conflict_by_area_order']


def resolve_loc_cue_conflict_by_priority(loc_cue: np.ndarray, priority: List[np.ndarray]) -> np.ndarray:
    """Solve conflict in candidate seed according to priority between channels.

    Args:
        loc_cue: (C, H, W) numpy array. Where (c, h, w) > 0 means channel c is one of the channel of the (h, w)
            pixel in sample n
        priority: Priority among channels. priority[index] = channels who prior to index'th channel

    Returns: (C, H, W) numpy array.
    """
    assert loc_cue.ndim == 3
    assert is_intarr(loc_cue)

    assert len(priority) == loc_cue.shape[0]

    one_hot_seed = loc_cue.copy()
    for ch_idx, prior_channels in enumerate(priority):
        one_hot_seed[ch_idx][np.sum(loc_cue[prior_channels, ...], axis=0) > 0] = 0

    assert np.all((np.sum(one_hot_seed, axis=0) <= 1) & np.sum(one_hot_seed, axis=0) >= 0)

    return one_hot_seed


def resolve_loc_cue_conflict_by_area_order(loc_cue_proposal, ignore_label, train_boat=False):
    """This function generate seed with priority strategy"""
    seed = np.ones(loc_cue_proposal.shape[:2], dtype=np.int64) * ignore_label

    # generate background seed
    seed[loc_cue_proposal[:, :, 0] == 1] = 0

    # generate foreground seed in the order of their area
    area = np.sum(loc_cue_proposal, axis=(0, 1))
    cls_order = np.argsort(area)[::-1]  # area in descending order
    for cls in cls_order:
        if area[cls] == 0:
            break
        seed[loc_cue_proposal[:, :, cls] == 1] = cls

    if train_boat:
        train_boat_ind = ((seed == 4) | (seed == 19)) & (loc_cue_proposal[:, :, 0] == 1)
        seed[train_boat_ind] = 0

    return seed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: balanced_seed_loss.py
@time: 2020/3/28 22:54
@desc:
"""
from typing import Tuple

import torch
import torch.nn.functional as F

__all__ = ['BalancedSeedloss']


class BalancedSeedloss(object):
    def __init__(self, eps: float=1e-5):
        """Compute balanced seed loss

        Args:
            eps: Min prob allowed when clamp probs
        """
        self.eps = eps

    def clamp_softmax(self, score, dim=1):
        probs = torch.clamp(F.softmax(score, dim), self.eps, 1)
        probs = probs / torch.sum(probs, dim=dim, keepdim=True)
        return probs

    def __call__(self, score_map, one_hot_label) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Compute balanced seed loss

        Args:
            score_map: (N, C, H, W) score map
            one_hot_label: (N, C, H, W) one-hot label

        Returns:
            Balanced Seed Loss
        """
        assert not one_hot_label.requires_grad
        pi = one_hot_label.to(torch.float)

        assert not torch.any(torch.isinf(score_map))
        assert not torch.any(torch.isnan(score_map))

        log_qi = torch.log(self.clamp_softmax(score_map))

        assert not torch.any(torch.isnan(log_qi))

        log_fg_qi = log_qi[:, 1:, :, :]
        fg_pi = pi[:, 1:, :, :]
        fg_count = torch.sum(fg_pi, dim=(1, 2, 3)) + self.eps

        log_bg_qi = log_qi[:, 0:1, :, :]
        bg_pi = pi[:, 0:1, :, :]
        bg_count = torch.sum(bg_pi, dim=(1, 2, 3)) + self.eps

        fg_loss_ = torch.sum(fg_pi * log_fg_qi, dim=(1, 2, 3))
        fg_loss = -1 * torch.mean(fg_loss_ / fg_count)  # mean reduce on batch

        bg_loss_ = torch.sum(bg_pi * log_bg_qi, dim=(1, 2, 3))
        bg_loss = -1 * torch.mean(bg_loss_ / bg_count)  # mean reduce on batch

        total_loss = bg_loss + fg_loss
        assert not torch.any(torch.isnan(total_loss)), \
            "fg_loss: {} fg_count: {} bg_loss: {} bg_count: {}".format(fg_loss, fg_count, bg_loss, bg_count)

        return total_loss, bg_loss, fg_loss



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/7/10 16:40
@File    : from_x_to_y.py
@Software: PyCharm
@Desc    : 
"""
import argparse

from alchemy_cat.py_tools import Config, gprint

parser = argparse.ArgumentParser()
parser.add_argument('--x', type=str, help='Source config path')
parser.add_argument('--y', type=str, help='Target config path')
parser.add_argument('--y_type', type=str, help='Target config type', default='alchemy-cat',
                    choices=['yaml', 'mmcv', 'alchemy-cat'])
args = parser.parse_args()

Config.from_x_to_y(args.x, args.y, y_type=args.y_type)
gprint(f"[SUCCESS] {args.x} ---------> {args.y} ({args.y_type})")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: polynomial_lr.py
@time: 2020/2/25 11:57
@desc:
"""
from typing import Any, Union

import warnings

from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer

from alchemy_cat.py_tools import indent

from .warnup_lr import WarmUpLR

__all__ = ['PolynomialLR', 'WarmPolynomialLR']


class PolynomialLR(_LRScheduler):
    def __init__(self, optimizer: Any, step_size: int, max_iter: int,
                 power: Union[int, float], last_epoch: int=-1):
        DeprecationWarning(f"{type(self)} is deprecated, use {WarmPolynomialLR} instead. ")
        self.step_size = step_size
        self.max_iter = max_iter
        self.power = power
        super(PolynomialLR, self).__init__(optimizer, last_epoch)

    def _polynomial_decay(self, lr):
        return lr * (1 - float(self.last_epoch) / self.max_iter) ** self.power

    def get_lr(self):
        if (
            (self.last_epoch == 0)
            or (self.last_epoch % self.step_size != 0)
            or (self.last_epoch > self.max_iter)
        ):
            return [group["lr"] for group in self.optimizer.param_groups]
        return [self._polynomial_decay(lr) for lr in self.base_lrs]

    def __repr__(self):
        return f"PolynomialLR <{self.__class__}>: \n" \
               + indent(f"step_size: {self.step_size}") + '\n' \
               + indent(f"max_iter: {self.max_iter}") + '\n' \
               + indent(f"power: {self.power}")


class WarmPolynomialLR(WarmUpLR):
    def __init__(self, optimizer: Optimizer, step_size: int=1, max_iter: int=10000,
                 power: int | float=.9, last_epoch: int=-1, min_lr: float=.0, end_lr_factors: float | list[float]=.0,
                 warm_up_iter: int=0, warm_up_min_lr: float=.0):
        if step_size != 1:
            DeprecationWarning(f"{step_size=} should be 1. ")

        assert 0 <= warm_up_iter < max_iter

        self.max_iter = max_iter
        self.power = power
        self.min_lr = min_lr
        self.end_lr_factors = (end_lr_factors if isinstance(end_lr_factors, list)
                               else [end_lr_factors] * len(optimizer.param_groups))
        super().__init__(optimizer, last_epoch, warm_up_iter, warm_up_min_lr)

    def _polynomial_decay(self, lr, end_lr_factor):
        return (lr - end_lr_factor * lr) * (1 - (self.last_epoch - self.warm_up_iter) /
                                            (self.max_iter - self.warm_up_iter)) ** self.power + end_lr_factor * lr

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        assert -1 <= self.last_epoch <= self.max_iter  # 该区间能给出合法学习率。

        if self.last_epoch < self.warm_up_iter:
            return WarmUpLR.get_lr(self)
        else:
            return [max(self._polynomial_decay(lr, end_lr_factor), self.min_lr)
                    for lr, end_lr_factor in zip(self.base_lrs, self.end_lr_factors, strict=True)]

    def __repr__(self):
        return f"PolynomialLR <{self.__class__}>: \n" \
               f"  max_iter: {self.max_iter}\n" \
               f"  power: {self.power}\n" \
               f"  min_lr: {self.min_lr}\n" \
               f"  end_lr: {self.end_lr_factors}\n" \
               f"  warm_up_iter: {self.warm_up_iter}\n" \
               f"  warm_up_min_lr: {self.warm_up_min_lr}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: random.py
@time: 2020/1/15 2:29
@desc:
"""
import hashlib
import random
from typing import Union

import numpy as np
try:
    import torch
    TORCH_AVAILABLE = True
except ImportError:
    torch = None
    TORCH_AVAILABLE = False

__all__ = ['hash_str', 'set_numpy_rand_seed', 'set_py_rand_seed', 'set_torch_rand_seed', 'set_rand_seed',
           'set_rand_seed_according_torch']


def hash_str(s: str, dynamic: bool=False) -> int:
    """Hash a str to int

    Args:
        s (str): str to be hashed
        dynamic (bool): Whether to use dynamic hash. If True, hash will be different in different processes

    Returns:
        int: Hashed int
    """
    if not dynamic:
        return int(hashlib.sha1(s.encode("utf-8")).hexdigest(), 16) % (2 ** 32)
    else:
        return hash(s) % (2 ** 32)


def set_numpy_rand_seed(seed: Union[int, str, np.ndarray]):
    """Set rand seed for numpy

    Args:
        seed (Union[int, str, np.ndarray]): int, str or np.ndarray seed, which will be hashed to get int seed
    """
    if isinstance(seed, str):
        seed = hash_str(seed, dynamic=False)
    elif isinstance(seed, int):
        seed = seed % (2 ** 32)
    elif not isinstance(seed, np.ndarray):
        raise ValueError(f"seed={seed} should be str, int or np.ndarray")

    np.random.seed(seed)


def set_torch_rand_seed(seed: Union[int, str]):
    """Set rand seed for torch on both cpu, cuda and cudnn

    Args:
        seed (Union[int, str]): int seed or str, which will be hashed to get int seed
    """
    if not TORCH_AVAILABLE:
        return
    if isinstance(seed, str):
        seed = hash_str(seed, dynamic=False)
    elif not isinstance(int(seed), int):
        raise ValueError(f"seed={seed} should be str or int")
    torch.manual_seed(int(seed))
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(int(seed))


def set_py_rand_seed(seed: Union[int, str]):
    """Set rand seed for python

    Args:
        seed (Union[int, str]): int seed or str, which will be hashed to get int seed
    """
    if isinstance(seed, str):
        seed = hash_str(seed, dynamic=False)
    elif not isinstance(int(seed), int):
        raise ValueError(f"seed={seed} should be str or int")
    random.seed(int(seed))


def set_rand_seed(seed: Union[int, str]):
    """Set rand seed for numpy, torch(cpu, cuda, cudnn), python

    Args:
        seed (Union[int, str]): int seed or str, which will be hashed to get int seed
    """
    set_numpy_rand_seed(seed)
    set_py_rand_seed(seed)
    set_torch_rand_seed(seed)


def set_rand_seed_according_torch():
    """Set rand seed according to torch process's rand seed
    The rand seed of non-torch libs may duplicate in several dataloader worker processes.
    Use this function as dataloader's worker init function can solve this problem.
    """
    if not TORCH_AVAILABLE:
        raise ImportError("torch is not available")
    seed = torch.initial_seed()
    set_py_rand_seed(seed)
    set_numpy_rand_seed(seed)



""" Input / Output module

Inputs / Outputs objects used to represent functions inputs / outputs
"""
from alchemy_cat.dag.errors import PyungoError


class _IO(object):
    """ IO Base class for inputs and outputs

    Args:
        name (str): The variable name of the input / output. Used when feeding input to node's func or find input's
            default value.
        map (str): Optional mapping name. Used when building graph.
        meta (dict): Not used yet.
        contract (str): Optional contract rule used by pycontracts.
    """

    def __init__(self, name, map=None, meta=None, contract=None):
        if not isinstance(name, str):
            raise PyungoError(f"IO name must be str, however get name = {name} with type {type(name)}")
        if '.' in name:
            raise PyungoError(f"IO can't have '.' in name, however get name = {name}")
        self._name = name

        self._meta = meta if meta is not None else {}
        self._value = None
        self._contract = None

        if map is not None:
            if not isinstance(map, str):
                raise PyungoError(f"IO map must be str, however get map = {map} with type {type(map)}")
            if '.' in map:
                raise PyungoError(f"IO can't have '.' in map, however get map = {map}")
        self._map = map if map is not None else self._name

        if contract:
            try:
                from contracts.main import parse_contract_string
            except ImportError:
                raise ImportError('pycontracts is needed to use contracts')
            self._contract = parse_contract_string(contract)

    @property
    def name(self):
        return self._name

    @property
    def map(self):
        return self._map

    @property
    def contract(self):
        return self._contract

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, x):
        """ When setting a value, we check the contract when applicable """
        if self._contract:
            self._contract.check(x)
        self._value = x


class Input(_IO):
    """ Placeholder a function's input

    Args:
        name (str): The variable name of the input / output. Used when feeding input to node's func or find input's
            default value.
        map (str): Optional mapping name. Used when building graph.
        is_kwarg (bool): Is kwarg input.
        value (Any): If not None, inp.value = value and is_constant = True.
        meta (dict): Not used yet.
        contract (str): Optional contract rule used by pycontracts.
    """

    def __init__(self, name, map=None, is_kwarg=False, value=None, meta=None, contract=None):
        super(Input, self).__init__(name, map, meta, contract)
        self.is_kwarg = is_kwarg
        self.is_constant = True if value is not None else False
        self.value = value

    def __repr__(self):
        return '<{} value={} is_kwarg: {} is_constant>'.format(
            self._name, self.value, self.is_kwarg, self.is_constant
        )


class Output(_IO):
    """ Placeholder of a function's output"""
    def __repr__(self):
        return '<{} value={}>'.format(
            self._name, self.value
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/9 22:12
@File    : loss_items.py
@Software: PyCharm
@Desc    : 
"""
import torch

__all__ = ['cal_loss_items']


def cal_loss_items(loss_items: dict, inp, out) -> dict:
    # * 损失字典，增加总损失一项。
    losses = {'total_loss': 0}

    # * 遍历计算每一个损失项，并汇入总损失。
    for item_name, loss_item in loss_items.items():
        # ** 跳过空loss项（一般是config中被取消了）。
        if not loss_item:
            continue
        # ** 计算损失。
        loss = loss_item.cal(loss_item.cri, inp, out)
        if torch.is_tensor(loss):  # 总是假设返回了多项损失。
            loss = (loss,)
        # ** 获取损失项的名字和权重。
        if loss_item.names:  # 如果指定了names，提取之。
            names, weights = loss_item.names, loss_item.weights
        else:  # 若没有指定names，说明loss_item本身就是loss名字。
            names, weights = (item_name,), (loss_item.weights,)

        for ls, name, weight in zip(loss, names, weights, strict=True):
            losses[name] = ls
            losses['total_loss'] += losses[name] * weight

    return losses



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: clamp_softmax.py
@time: 2020/3/28 23:02
@desc:
"""
import numpy as np
import torch
import torch.nn.functional as F

__all__ = ['clamp_probs', 'clamp_softmax', 'clip_probs']


def clamp_probs(probs: torch.Tensor, eps: float=1e-5) -> torch.Tensor:
    """Clamp and normalize probs to make sure that min prob in probs won't be too small

    Args:
        probs: (N, C, H, W) probs
        eps: Min prob allowed when clamp probs

    Returns:
        Clamped and normalized probs with shape (N, C, H, W)
    """
    clamped_probs = torch.clamp(probs, eps, 1)
    normalized_probs = clamped_probs / torch.sum(clamped_probs, dim=1, keepdim=True)
    return normalized_probs


def clamp_softmax(score_map: torch.Tensor, eps: float=1e-5):
    """Calculate probs from score map, then clamp and normalize probs to make sure that min prob in probs won't be
    too small

    Args:
        score_map: (N, C, H, W) score map
        eps: Min prob allowed when clamp probs

    Returns:
        Clamped and normalized probs with shape (N, C, H, W)
    """
    return clamp_probs(F.softmax(score_map, 1), eps)

def clip_probs(probs: np.ndarray, eps: float=1e-5) -> np.ndarray:
    """Clip and normalize probs to make sure that min prob in probs won't be too small

    Args:
        probs: (N, C, H, W) probs
        eps: Min prob allowed when clip probs

    Returns:
        Clip and normalized probs with shape (N, C, H, W)
    """
    clipped_probs = np.clip(probs, eps, 1)
    normalized_probs = clipped_probs / np.sum(clipped_probs, axis=1, keepdims=True)
    return normalized_probs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/10 13:51
@File    : parse_log.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional, Callable
from dataclasses import dataclass

import re

import numpy as np

__all__ = ['Field', 'find_fields_in_log']


@dataclass
class Field(object):
    """字段如何被查找和解析。"""
    pattern: Optional[str] = None
    parser: Callable = float
    val: Optional[list] = None


def find_fields_in_log(log_file: str, fields: dict, np_val: bool=False):
    """在日志中找到指定字段的值。

    Args:
        log_file: 日志文件。
        fields: 字段。
        np_val: 匹配到的字段是否要转为nparray，再赋值给字段的val。
    """
    # * 读取要解析的训练日志。
    with open(log_file, 'r', encoding='utf-8') as f:
        log = f.read()
    # * 循环匹配，获取字段。
    for field in fields.values():
        field.val = [field.parser(v) for v in re.findall(field.pattern, log)]
        if np_val:
            field.val = np.array(field.val)



# -- configs/tune/tune_bs_epoch,pri,name/cfg.py --

from alchemy_cat.dl_config import Cfg2Tune, Param2Tune

cfg = Cfg2Tune(caps='configs/mnist/base,sched_from_addon/cfg.py')

cfg.loader.ini.batch_size = Param2Tune([128, 256, 512], optional_value_names=['1xbs', '2xbs', '4xbs'], priority=1)

cfg.sched.epochs = Param2Tune([5, 15], priority=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/25 20:18
@File    : cfg.py
@Software: PyCharm
@Desc    : 
"""
from numpy import sum, min, max
from alchemy_cat.py_tools.param_tuner.tests.configs.config_lib import func_a, func_b, func_c, func_d

from alchemy_cat.py_tools import Cfg2Tune, Param2Tune


cfg = config = Cfg2Tune()

cfg.rslt_dir = 'imported_outside_config'

# * 外部库中函数。
cfg.foo0.foo1.a = Param2Tune([sum, min, max])
# * 当前项目中函数。
cfg.foo0.foo1.b = Param2Tune([func_a, func_b, func_c, func_d])
# * Param on diff Cfg2Tune
cfg.foo0.fix = Param2Tune(['foo.bar'])
# * Param on root Cfg2Tune
cfg.c = Param2Tune([False, True])
# * 静态值，等于外部库函数或项目中函数。
cfg.foo0.a_max = sum
cfg.foo0.a_min = (func_c, func_d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: type.py
@time: 2020/1/8 8:13
@desc:
"""
from typing import Any

import numpy as np
import torch


__all__ = ["is_intarr", "is_int", "is_floatarr", "is_float", "is_boolarr", "is_bool",
           "tolist", "totuple", "dict_with_arr_is_eq"]


def is_int(elem) -> bool:
    """
    Args:
        elem (UnKnown): element need to be judged. Note in this func, bool is not int.

    Returns:
        Is the elem is an int type(int, np.int*, torch.int*)
    """
    if isinstance(elem, torch.Tensor):
        if elem.ndim != 0:
            return False
        elif (not torch.is_floating_point(elem)) and (not torch.is_complex(elem)) and (elem.dtype is not torch.bool):
            return True
        else:
            return False

    if isinstance(elem, np.integer):
        return True

    if isinstance(elem, int) and (not isinstance(elem, bool)):
        return True

    return False


def is_intarr(arr) -> bool:
    """
    Args:
        arr (UnKnown): arr need to be judged. Note for torch, tensor(1) is both int and intarr.

    Returns:
        Is the arr is an int type(np.int*, torch.int*)
    """
    if (isinstance(arr, torch.Tensor) and (not torch.is_floating_point(arr)) and (not torch.is_complex(arr))
            and (arr.dtype is not torch.bool)
            and arr.ndim > 0):
        return True

    if isinstance(arr, np.ndarray) and np.issubdtype(arr.dtype, np.integer):
        return True

    return False


def is_float(elem) -> bool:
    """
    Args:
        elem (UnKnown): element need to be judged

    Returns:
        Is the elem is a float type(float, np.float*, torch.float*)
    """
    if isinstance(elem, torch.Tensor):
        if elem.ndim != 0:
            return False
        elif torch.is_floating_point(elem):
            return True
        else:
            return False

    if isinstance(elem, np.floating):
        return True

    if isinstance(elem, float):
        return True

    return False


def is_floatarr(arr) -> bool:
    """
    Args:
        arr (UnKnown): arr need to be judged.

    Returns:
        Is the arr is a float type(np.float*, torch.float*)
    """
    if isinstance(arr, torch.Tensor) and torch.is_floating_point(arr) and arr.ndim > 0:
        return True

    if isinstance(arr, np.ndarray) and np.issubdtype(arr.dtype, np.floating):
        return True

    return False


def is_bool(elem) -> bool:
    """
    Args:
        elem (UnKnown): element need to be judged

    Returns:
        Is the elem is a bool type(int, np.int*, torch.int*)
    """
    if isinstance(elem, torch.Tensor):
        if elem.ndim != 0:
            return False
        elif elem.dtype is torch.bool:
            return True
        else:
            return False

    if isinstance(elem, np.bool_):
        return True

    if isinstance(elem, bool):
        return True

    return False


def is_boolarr(arr) -> bool:
    """
    Args:
        arr (UnKnown): arr need to be judged.

    Returns:
        Is the arr is a bool type(np.float*, torch.float*)
    """
    if isinstance(arr, torch.Tensor) and (arr.dtype is torch.bool) and arr.ndim > 0:
        return True

    if isinstance(arr, np.ndarray) and (arr.dtype == np.bool_):
        return True

    return False


def tolist(arr: Any) -> list:
    """Return arr.tolist() for ndarray and tensor, other wise return list(arr)
        list(tensor) or list(ndarray) return a list with tensor or numpy scalar when arr with dim=1, which
        may cause confusing result for some function can't recognize tensor or scalar as python datatype.

        When the arr to be converted to list can be tensor/ndarray or other type both, this function can make sure
        the output list has python datatype element.
    Args:
        arr: arr need to be converted to list

    Returns:
        converted arr with type list
    """
    if isinstance(arr, torch.Tensor) or isinstance(arr, np.ndarray):
        return arr.tolist()
    else:
        return list(arr)


def totuple(arr: Any) -> tuple:
    """Return tuple(arr.tolist()) for ndarray and tensor, other wise return list(arr)
        tuple(tensor) or tuple(ndarray) return a tuple with tensor or numpy scalar when arr with dim=1, which
        may cause confusing result for some function can't recognize tensor or scalar as python datatype.

        When the arr to be converted to tuple can be tensor/ndarray or other type both, this function can make sure
        the output tuple has python datatype element.
    Args:
        arr: arr need to be converted to tuple

    Returns:
        converted arr with type tuple
    """
    if isinstance(arr, torch.Tensor) or isinstance(arr, np.ndarray):
        return tuple(arr.tolist())
    else:
        return tuple(arr)


def dict_with_arr_is_eq(dict1: dict, dict2: dict, request_order: bool=False, rtol: float=0.0, atol: float=0.0):
    """Compare to dict is equal when some dict's values is Tensor or ndarray

    Args:
        dict1: Fist dict to be compared.
        dict2: Second dict to be compared
        request_order: If True, orders of dict1 and dict2 will be compared. (Default: True)
        rtol: param rtol for np.isclose() when compare two ndarray values
        atol: param atol for np.isclose() when compare two ndarray values

    Returns:
        Is two dicts have keys and values
    """

    if len(dict1) != len(dict2):
        return False

    if request_order:
        if dict1.keys() != dict2.keys():
            return False

    for key1, val1 in dict1.items():
        if key1 not in dict2:
            return False

        val2 = dict2[key1]

        val1 = val1.detach().cpu().numpy() if isinstance(val1, torch.Tensor) else val1
        val2 = val2.detach().cpu().numpy() if isinstance(val2, torch.Tensor) else val2

        if type(val1) is not type(val2):
            return False

        if isinstance(val1, np.ndarray):
            if not np.all(np.isclose(val1, val2, rtol, atol)):
                return False
        else:
            if val1 != val2:
                return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/5 21:08
@File    : few_shot_dt.py
@Software: PyCharm
@Desc    : 
"""
from itertools import chain

import numpy as np
import torch
from alchemy_cat.data import Dataset

from .typing import SegDataset

__all__ = ['FewShotDt']


class FewShotDt(Dataset):
    def __init__(self, ori_dt: SegDataset, shot_num: int, seed: int, except_bg: bool=False):
        assert hasattr(ori_dt, 'get_by_img_id')
        assert hasattr(ori_dt, 'image_ids')
        assert hasattr(ori_dt, 'id2cls_labels')

        self.ori_dt = ori_dt
        self.shot_num = shot_num
        self.seed = seed
        self.except_bt = except_bg

        self.class_names = ori_dt.class_names
        self.class_num = ori_dt.class_num
        self.ignore_label = ori_dt.ignore_label

        self.cls_image_ids = {}
        self.image_ids = []

        self._sample_shots()

    def _sample_shots(self):
        # * 获得img_id - cls_label关系表。img_id来自原数据集。
        cls_labels = np.stack((self.ori_dt.id2cls_labels[img_id] for img_id in self.ori_dt.image_ids), axis=0)
        cls_labels = (cls_labels == 1).astype(np.int32)  # 过滤掉ignore=255标签，int32类型防止sum后溢出。
        cls_labels = torch.from_numpy(cls_labels)
        cls_num = cls_labels.shape[1]
        # * 对每个类别（可能要跳过背景）分别采样shot_num个样本。
        for cls in range(cls_num):
            # * （可选）跳过背景。
            if cls == 0 and self.except_bt:
                continue
            # * 在cls类别采样shot_num个img_id。
            labels = cls_labels[:, cls]
            shot_indices = torch.multinomial(labels.to(torch.float32), min(self.shot_num, labels.sum()),
                                             replacement=False,
                                             generator=torch.Generator('cpu').manual_seed(self.seed + cls))  # 随机隔离。
            self.cls_image_ids[cls] = [self.ori_dt.image_ids[idx] for idx in shot_indices]
            # * 防止重复采样。
            cls_labels[shot_indices, :] = 0
        # * 拼接得到image_ids。
        self.image_ids = list(chain(*self.cls_image_ids.values()))
        assert len(set(self.image_ids)) == len(self.image_ids)  # 确保无重复采样。

    def get_item(self, index: int):
        return self.ori_dt.get_by_img_id(self.image_ids[index])

    def __len__(self):
        return len(self.image_ids)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/6/20 22:00
@File    : slide_inference.py
@Software: PyCharm
@Desc    : 滑动窗口法推理图片。
"""
from typing import Callable, Union, Iterable, Tuple, List

from math import ceil

import torch
import torch.nn.functional as F

from alchemy_cat.alg import size2HW, find_nearest_odd_size

__all__ = ['slide_inference', 'bisection']


def bisection(n: int):
    return n // 2, n // 2 + n % 2


def _pad_imgs(imgs: torch.Tensor, padded_h: int, padded_w: int, pad_val: ...=0) -> Tuple[torch.Tensor, List[int]]:
    pad_top, pad_bottom = bisection(max(padded_h - imgs.shape[-2], 0))
    pad_left, pad_right = bisection(max(padded_w - imgs.shape[-1], 0))

    return (F.pad(imgs, [pad_left, pad_right, pad_top, pad_bottom], mode='constant', value=pad_val),
            [pad_top, pad_bottom, pad_left, pad_right])


def slide_inference(imgs: torch.Tensor, *others: torch.Tensor,
                    model: Callable[[torch.Tensor, ...], torch.Tensor],
                    window_sizes: Union[Iterable, int], strides: Union[Iterable, int],
                    num_class: int=None,
                    pad: bool=False,
                    win_size_checker: Callable[[int], bool]=lambda x: find_nearest_odd_size(x, min_n=4) == x,
                    align_corners: bool=True)\
        -> torch.Tensor:
    """滑动窗口法推理图片。

    Args:
        imgs: 待推理的(N, 3, H, W)图片。
        others: 其他输入，如标签，形状为(..., H, W)。
        model: 推理函数，输入(N, 3, H, W)图片，输出得分。
        window_sizes: 滑动窗口尺寸。若为可迭代对象，表(win_h, win_w)；若为int，则正方形窗口的边长。
        num_class: 模型logit输出通道数。若为None，则使用模型的输出通道数。
        strides: 滑动窗口步长。若为可迭代对象，表(stride_h, stride_w)；若为int，表步长之于高宽上一致者。
        pad: 是否在推理前，将输入图片填充到滑动窗口尺寸。注意，若没有pad，当确保输入图片尺寸为模型支持的尺寸。
        win_size_checker: 检查最后真实切出的窗口，其尺寸是否符合要求。默认检查是否为n=4的奇尺寸。
        align_corners: 是否对齐角点。

    Returns:
        与输入imgs等大的(N, num_class, H, W)得分。
    """
    # * others的形状必须与imgs的形状一致。
    assert all([imgs.shape[2:] == other.shape[-2:] for other in others])

    # * 解析输入参数。
    win_size_h, win_size_w = size2HW(window_sizes)
    stride_h, stride_w = size2HW(strides)

    # * 填充图片。
    if pad:
        padded_imgs, margin = _pad_imgs(imgs, win_size_h, win_size_w)
        padded_others = [_pad_imgs(other, win_size_h, win_size_w)[0] for other in others]
    else:
        padded_imgs, margin = imgs, [0] * 4
        padded_others = others

    # * 初始化得分和累加次数。
    def ini_logits(c: int) -> torch.Tensor:
        return torch.zeros((padded_imgs.shape[0], c, padded_imgs.shape[2], padded_imgs.shape[3]),
                           dtype=padded_imgs.dtype, device=padded_imgs.device)

    logits = ini_logits(num_class) if num_class is not None else None
    add_count = torch.zeros((padded_imgs.shape[2], padded_imgs.shape[3]),
                            dtype=torch.int64, device=padded_imgs.device)

    # * 计算滑动窗口。
    h_grids = ceil(max(padded_imgs.shape[2] - win_size_h, 0) / stride_h) + 1
    w_grids = ceil(max(padded_imgs.shape[3] - win_size_w, 0) / stride_w) + 1

    # * 逐窗口推理、累加。
    for h_idx in range(h_grids):
        y1 = h_idx * stride_h
        y2 = min(y1 + win_size_h, padded_imgs.shape[2])
        y1 = max(y2 - win_size_h, 0)
        for w_idx in range(w_grids):
            x1 = w_idx * stride_w
            x2 = min(x1 + win_size_w, padded_imgs.shape[3])
            x1 = max(x2 - win_size_w, 0)
            cropped_imgs = padded_imgs[:, :, y1:y2, x1:x2]
            cropped_others = [other[..., y1:y2, x1:x2] for other in padded_others]

            # assert find_nearest_odd_size(cropped_imgs.shape[2], min_n=4) == cropped_imgs.shape[2]
            # assert find_nearest_odd_size(cropped_imgs.shape[3], min_n=4) == cropped_imgs.shape[3]
            assert win_size_checker(cropped_imgs.shape[2])
            assert win_size_checker(cropped_imgs.shape[3])

            cropped_logits = F.interpolate(model(cropped_imgs, *cropped_others),
                                           size=(cropped_imgs.shape[2], cropped_imgs.shape[3]),
                                           mode='bilinear', align_corners=align_corners)
            if logits is None:
                logits = ini_logits(cropped_logits.shape[1])
            logits[:, :, y1:y2, x1:x2] += cropped_logits
            add_count[y1:y2, x1:x2] += 1

    # * 求平均得分。
    assert torch.all(add_count != 0)
    logits /= add_count

    # * 后处理得到与输入等大的得分图。
    return logits[:, :, margin[0]:(logits.shape[2] - margin[1]), margin[2]:(logits.shape[3] - margin[3])]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/4/13 23:15
@File    : seg_criterion.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional

import torch
from torch import nn
import torch.nn.functional as F

__all__ = ['UpsampleCE']


class UpsampleCE(nn.Module):

    def __init__(self, ignore_label: int=255, weight: Optional[torch.Tensor]=None):
        """给定ignore label和类别权重。前向时将logit上采样到label尺寸，计算交叉熵损失。"""
        super().__init__()
        self.criterion = nn.CrossEntropyLoss(reduction='mean', ignore_index=ignore_label, weight=weight)

    def forward(self, logit, label):
        if logit.shape[2:] != label.shape[1:]:
            upsample_logit = F.interpolate(logit, size=label.shape[1:], mode='bilinear', align_corners=True)
        else:
            upsample_logit = logit
        return self.criterion(upsample_logit, label)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: color_print.py
@time: 2021/9/13 1:27
@desc:
"""
from colorama import Fore, Style

__all__ = ['rprint', 'bprint', 'yprint', 'cprint', 'gprint']


def rprint(*args, **kwargs):
    print(Fore.RED, end='')
    print(*args, **kwargs)
    print(Style.RESET_ALL, end='')


def bprint(*args, **kwargs):
    print(Fore.BLUE, end='')
    print(*args, **kwargs)
    print(Style.RESET_ALL, end='')


def yprint(*args, **kwargs):
    print(Fore.YELLOW, end='')
    print(*args, **kwargs)
    print(Style.RESET_ALL, end='')


def cprint(*args, **kwargs):
    print(Fore.CYAN, end='')
    print(*args, **kwargs)
    print(Style.RESET_ALL, end='')


def gprint(*args, **kwargs):
    print(Fore.GREEN, end='')
    print(*args, **kwargs)
    print(Style.RESET_ALL, end='')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/23 19:52
@File    : score2seed.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional, Callable

import numpy as np

from .utils import get_bg_fg_score, cls_in_label2tag_always_bg, complement_entropy_with_shift
from .loc_cue import get_loc_cue_max_thresh, get_loc_cue_contrib_thresh
from .solve_conflict import solve_conflict_by_area, solve_conflict_on_channel

__all__ = ["cw_pred_entropy_thresh", "sw_pred_max_thresh_solve_area", "sw_pred_contrib_thresh_solve_channel",
           "sw_pred_contrib_thresh_solve_area", "sw_pred_max_thresh_solve_channel"]


def cw_pred_entropy_thresh(score: np.ndarray, cls_in_label: np.ndarray, alpha: float = 6.,
                           thresh: float = .05, shift: float = .1, ignore_label: int = 255,
                           score_refiner: Callable[[np.ndarray], np.ndarray]=lambda x: x) -> np.ndarray:
    bg_fg_score = score_refiner(get_bg_fg_score(score, alpha))

    pred = np.ones(score.shape[1:], dtype=np.int64) * ignore_label

    comp_entropy = complement_entropy_with_shift(bg_fg_score[None, ...], shift)[0]
    confident_loc = comp_entropy > thresh

    tag = cls_in_label2tag_always_bg(cls_in_label)

    pred[confident_loc] = tag[np.argmax(bg_fg_score, axis=0), ][confident_loc]

    return pred


def sw_pred_max_thresh_solve_area(score: np.ndarray, cls_in_label: np.ndarray,
                                  thresh: float = .2, alpha: float = 6.,
                                  area: Optional[np.ndarray] = None, ignore_label: int = 255, train_boat: bool = True,
                                  score_refiner: Callable[[np.ndarray], np.ndarray]=lambda x: x) -> np.ndarray:
    bg_fg_score = score_refiner(get_bg_fg_score(score, alpha))
    loc_cue = get_loc_cue_max_thresh(bg_fg_score, thresh)
    return solve_conflict_by_area(loc_cue, cls_in_label, area, ignore_label, train_boat)


def sw_pred_max_thresh_solve_channel(score: np.ndarray, cls_in_label: np.ndarray,
                                     s_thresh: float = .2, s_alpha: float = 6.,
                                     c_alpha: float = 6., c_thresh: float = .05, shift: float = .1,
                                     ignore_label: int = 255,
                                     score_refiner: Callable[[np.ndarray], np.ndarray]=lambda x: x) -> np.ndarray:
    cw_pred = cw_pred_entropy_thresh(score, cls_in_label, c_alpha, c_thresh, shift, ignore_label, score_refiner)

    bg_fg_score = score_refiner(get_bg_fg_score(score, s_alpha))
    loc_cue = get_loc_cue_max_thresh(bg_fg_score, s_thresh)
    return solve_conflict_on_channel(loc_cue, cls_in_label, cw_pred, ignore_label)


def sw_pred_contrib_thresh_solve_area(score: np.ndarray, cls_in_label: np.ndarray,
                                      thresh: float = .8, alpha: float = 6.,
                                      area: Optional[np.ndarray] = None, ignore_label: int = 255,
                                      train_boat: bool = True,
                                      score_refiner: Callable[[np.ndarray], np.ndarray]=lambda x: x) -> np.ndarray:
    bg_fg_score = score_refiner(get_bg_fg_score(score, alpha))
    loc_cue = get_loc_cue_contrib_thresh(bg_fg_score, thresh)
    return solve_conflict_by_area(loc_cue, cls_in_label, area, ignore_label, train_boat)


def sw_pred_contrib_thresh_solve_channel(score: np.ndarray, cls_in_label: np.ndarray,
                                         s_thresh: float = .8, s_alpha: float = 6.,
                                         c_alpha: float = 6., c_thresh: float = .05, shift: float = .1,
                                         ignore_label: int = 255,
                                         score_refiner: Callable[[np.ndarray], np.ndarray]=lambda x: x) -> np.ndarray:
    cw_pred = cw_pred_entropy_thresh(score, cls_in_label, c_alpha, c_thresh, shift, ignore_label, score_refiner)

    bg_fg_score = score_refiner(get_bg_fg_score(score, s_alpha))
    loc_cue = get_loc_cue_contrib_thresh(bg_fg_score, s_thresh)
    return solve_conflict_on_channel(loc_cue, cls_in_label, cw_pred, ignore_label)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: msc_flip_inference.py
@time: 2020/1/16 7:38
@desc:
"""
from collections import abc
from typing import Union, Callable, Iterable

import torch
import torch.nn.functional as F

from alchemy_cat.alg.utils import size2HW

__all__ = ['msc_flip_inference', 'flip_imgs', 'merge_flipped_probs']


def bisection(n: int):
    return n // 2, n // 2 + n % 2


def tensor_pad_imgs(imgs: torch.Tensor, size, value: float=0.) -> tuple[torch.Tensor, list[int]]:
    """中心填充imgs到size，若图片某个尺度大于size，则该方向不填充。

    Args:
        imgs: (N, C, H, W)图片张量。
        size: 要pad到的大小，size: int表示pad到方形size，(h: int, w: int)表示要填充到的高、宽。
        value: 要填充的值，默认为0。

    Returns:
        (填充后的张量，上、下、左、右的填充尺寸)。
    """
    padded_h, padded_w = size2HW(size)

    pad_top, pad_bottom = bisection(max(padded_h - imgs.shape[2], 0))
    pad_left, pad_right = bisection(max(padded_w - imgs.shape[3], 0))

    return F.pad(imgs, [pad_left, pad_right, pad_top, pad_bottom], mode='constant', value=value), [pad_top, pad_bottom,
                                                                                                   pad_left, pad_right]

# def _pad_imgs(imgs, size):
#     img_h, img_w = imgs.size(2), imgs.size(3)
#     padded_h, padded_w = size2HW(size)
#
#     pad_h = padded_h - img_h
#     pad_w = padded_w - img_w
#
#     if pad_h < 0 or pad_w < 0:
#         raise ValueError(f"pad_h={pad_h} or pad_w={pad_w} should equal or larger than 0")
#
#     return F.pad(imgs, [0, pad_w, 0, pad_h], mode='constant', value=0.0)


# def _pad_labels(labels, size, ignore_label):
#     label_h, label_w = labels.size(1), labels.size(2)
#     padded_h, padded_w = size2HW(size)
#
#     pad_h = padded_h - label_h
#     pad_w = padded_w - label_w
#
#     if pad_h < 0 or pad_w < 0:
#         raise ValueError(f"pad_h={pad_h} or pad_w={pad_w} should equal or larger than 0")
#
#     return F.pad(labels, [0, pad_w, 0, pad_h], mode='constant', value=ignore_label)

def flip_imgs(imgs):
    """Flip imgs and cat it N dim"""
    return torch.cat([imgs, torch.flip(imgs, [3])], dim=0)


def merge_flipped_probs(probs):
    """merge flipped probs by calculate the average of origin and flipped prob"""
    dual_logits = probs.view(2, probs.shape[0] // 2, *probs.shape[1:])
    dual_logits[1] = dual_logits[1].flip([3])
    return torch.mean(dual_logits, dim=0)


def msc_flip_inference(imgs: torch.Tensor, model: Callable[[torch.Tensor], torch.Tensor],
                       msc_factors: Iterable[Union[Iterable[float], float]],
                       is_flip: bool=True,
                       pad_imgs_to: Union[None, Iterable, int]=None,
                       pad_aligner: Union[Callable[[int], int], Iterable[Callable[[int], int]]]=lambda x: x,
                       msc_aligner: Union[Callable[[int], int], Iterable[Callable[[int], int]]]=lambda x: x,
                       align_corners: bool=False,
                       cuda_memory_saving: int=0,
                       softmax_norm: bool=True) \
        -> torch.Tensor:
    """MSC and flip inference

    Args:
        imgs: Imgs with shape (N, C, H, W)
        model: torch models
        msc_factors: Multi scale factors. Each factor can be 'factor_for_height_width' or
            '(height_factor, width_factor)'.
        is_flip: If true, the flipped imgs will be used for inference
        pad_imgs_to: If not None, the img will be padded to size(int or [int, int]) specified
        pad_aligner: The pad_size will be fixed by aligner(pad_size). If Iterable, then first and second aligners
            separately used to align H and W.
        msc_aligner: The scaled_size calculated by scale_factor * img_size will be fixed by aligner(scaled_size). If
            Iterable, then first and second aligners separately used to align H and W.
        align_corners: If True, use `align_corners` parameter for torch.nn.functional.interpolate.
        cuda_memory_saving: int before 0-2. The larger less_cuda_memory set, the less gpu memory used by function. May
            loss some performance. (Default: 0)
        softmax_norm: If Ture, the output logit will be normalized by softmax then fusion between different scales. Else
            the output logit will be implemented as prob and directly fusion and output.

    Returns: (N, C, H, W) probs of image predicts
    """
    origin_h, origin_w = imgs.shape[-2:]

    # * For saving more gpu memory, move imgs to cpu.
    if cuda_memory_saving > 1:
        imgs = imgs.cpu()

    # * Process pad aligner
    if isinstance(pad_aligner, abc.Callable):
        pad_aligner_h, pad_aligner_w = pad_aligner, pad_aligner
    elif isinstance(pad_aligner, abc.Iterable):
        pad_aligner_h, pad_aligner_w = pad_aligner
    else:
        raise ValueError(f"pad_aligner {pad_aligner} must be Callable or Iterable[Callable]")

    # * Process scale aligner
    if isinstance(msc_aligner, abc.Callable):
        msc_aligner_h, msc_aligner_w = msc_aligner, msc_aligner
    elif isinstance(msc_aligner, abc.Iterable):
        msc_aligner_h, msc_aligner_w = msc_aligner
    else:
        raise ValueError(f"msc_aligner {msc_aligner} must be Callable or Iterable[Callable]")

    # * Get probs from each scale
    msc_probs = []
    for msc_factor in msc_factors:
        # * Process msc factor
        if isinstance(msc_factor, float):
            msc_factor_h, msc_factor_w = msc_factor, msc_factor
        elif isinstance(msc_factor, abc.Iterable):
            msc_factor_h, msc_factor_w = msc_factor
        else:
            raise ValueError(f"msc_factor {msc_factor} in msc_factors {msc_factors} must be float or Iterable[float]")

        # * Get scaled size
        scaled_h, scaled_w = round(origin_h * msc_factor_h), round(origin_w * msc_factor_w)
        scaled_h, scaled_w = msc_aligner_h(scaled_h), msc_aligner_w(scaled_w)

        batch_X = F.interpolate(imgs, size=(scaled_h, scaled_w), mode='bilinear', align_corners=align_corners)
        if is_flip:
            batch_X = flip_imgs(batch_X)

        # * pad img to align(pad_img_to) size
        pad_imgs_to_h, pad_imgs_to_w = size2HW(pad_imgs_to) if pad_imgs_to is not None else (scaled_h, scaled_w)
        pad_align_size = (pad_aligner_h(pad_imgs_to_h), pad_aligner_w(pad_imgs_to_w))
        padded_batch_X, margin = tensor_pad_imgs(batch_X, pad_align_size)
        padded_h, padded_w = padded_batch_X.shape[-2:]

        padded_logits = model(padded_batch_X.cuda())

        if cuda_memory_saving > 0:
            padded_logits = padded_logits.cpu()

        if margin == [0, 0, 0, 0]:
            logits = F.interpolate(padded_logits, size=(origin_h, origin_w),
                                   mode='bilinear', align_corners=align_corners)
        else:
            logits = F.interpolate(padded_logits, size=(padded_h, padded_w),
                                   mode='bilinear', align_corners=align_corners)
            logits = logits[:, :, margin[0]:(logits.shape[2] - margin[1]), margin[2]:(logits.shape[3] - margin[3])]
            logits = F.interpolate(logits, size=(origin_h, origin_w),
                                   mode='bilinear', align_corners=align_corners)
        probs = F.softmax(logits, dim=1) if softmax_norm else logits
        if is_flip:
            probs = merge_flipped_probs(probs)

        msc_probs.append(probs)
    # * Get probs with padded size
    probs = torch.stack(msc_probs, dim=0).mean(dim=0)

    # * return with size of origin img
    return probs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: my_logger.py
@time: 2019/7/29 12:18
@desc:
"""
import sys


class Const(object):
    class ConstError(TypeError):
        pass

    class ConstCaseError(ConstError):
        pass

    def __setattr__(self, name, value):
        if name in self.__dict__:
            raise self.ConstError("Can't change const.%s" % name)

        object.__setattr__(self, name, value)

    def __delattr__(self, name):
        if name in self.__dict__:
            raise self.ConstError("can't unbind const(%s)" % name)
        raise NameError(name)


# sys.modules[__name__] = Const()
# K = Const()



# -- configs/mnist/plain_usage,override_loader/cfg.py --

from alchemy_cat.dl_config import Config

cfg = Config(caps='configs/mnist/plain_usage/cfg.py')

cfg.loader.ini.override()
cfg.loader.ini.shuffle = False
cfg.loader.ini.drop_last = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/7 10:12
@File    : test_param2tune.py
@Software: PyCharm
@Desc    :
"""
import pytest

import sys
sys.path = ['', 'py_tools/param_tuner/tests'] + sys.path  # noqa: E402

from alchemy_cat.py_tools import Param2Tune


@pytest.fixture(scope="function")
def param():
    def param_subject_to(opt_val):
        return opt_val % 2 == 0
    return Param2Tune([0, 1, 2, 3], param_subject_to)


def test_reset(param):
    assert param._cur_val is None
    next(iter(param))
    assert param._cur_val == 0
    param.reset()
    assert param._cur_val is None


def test_none_cur_val(param):
    param.reset()
    with pytest.raises(RuntimeError, match="Check if the former param call later param"):
        _ = param.cur_val


def test_iter(param):
    iter1_param = iter(param)
    assert (0, 0) == next(iter1_param) == (param.cur_val, param.cur_val_name)
    assert (2, 2) == next(iter1_param) == (param.cur_val, param.cur_val_name)
    with pytest.raises(StopIteration):
        next(iter1_param)
    assert param._cur_val is None and param._cur_val_name is None

    iter2_param = iter(param)
    assert (0, 0) == next(iter2_param) == (param.cur_val, param.cur_val_name)
    assert (2, 2) == next(iter2_param) == (param.cur_val, param.cur_val_name)
    with pytest.raises(StopIteration):
        next(iter2_param)
    assert param._cur_val is None and param._cur_val_name is None


def test_subject_to(param):
    legal_val = list(param)
    assert legal_val == [(0, 0), (2, 2)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2020/11/6 10:26
@File    : intermediate_value_getter.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional

from torch import nn
import torch

__all__ = ["IntermediateValueGetter"]


class IntermediateValueGetter(object):
    """Get intermediate value of torch module"""
    def __init__(self, module: nn.Module, get_forward: bool=False, get_backward: bool=False):
        """Get intermediate value of torch module

        Args:
            module: Module whose intermediate value need be gotten.
            get_forward: Is get forward input/output (Default: False).
            get_backward: Is get backward grad_input/grad_output (Default: False).
        """
        self.get_forward, self.get_backward = get_forward, get_backward

        self._module = None
        self._input, self._output, self._grad_input, self._grad_output = None, None, None, None

        self._forward_hook_handler: Optional[torch.utils.hooks.RemovableHandle] = None
        self._backward_hook_handler: Optional[torch.utils.hooks.RemovableHandle] = None

        self.module = module

    def _forward_hook_fn(self, _, input, output):
        self._input = input
        self._output = output

    def _backward_hook_fn(self, _, grad_input, grad_output):
        self._grad_input = grad_input
        self._grad_output = grad_output

    def _set_hooks(self, module: nn.Module):
        if self.get_forward:
            self._forward_hook_handler = module.register_forward_hook(self._forward_hook_fn)
        if self.get_backward:
            self._backward_hook_handler = module.register_backward_hook(self._backward_hook_fn)

    @property
    def module(self):
        return self._module

    @module.setter
    def module(self, module):
        self.remove_hooks()
        self._set_hooks(module)
        self._module = module

    def remove_hooks(self):
        if self._forward_hook_handler is not None:
            self._forward_hook_handler.remove()
        if self._backward_hook_handler is not None:
            self._backward_hook_handler.remove()

    def __del__(self):
        self.remove_hooks()

    @property
    def input(self):
        assert self.get_forward
        return self._input

    @property
    def output(self):
        assert self.get_forward
        return self._output

    @property
    def grad_input(self):
        assert self.get_backward
        return self._grad_input

    @property
    def grad_output(self):
        assert self.get_backward
        return self._grad_output

    def __repr__(self):
        return f"IntermediateValueGetter: {self.__class__.__name__}\n" \
                + f"    module: {self.module}\n" \
                + f"    get_forward: {self.get_forward}\n" \
                + f"    get_backward: {self.get_backward}\n" \
                + f"    forward_hook_handler: {self._forward_hook_handler}\n" \
                + f"    backward_hook_handler: {self._backward_hook_handler}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/4/17 15:51
@File    : complement_entropy_confidence.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np
from scipy.stats import entropy

__all__ = ['complement_entropy_confidence']


def complement_entropy_confidence(prob: np.ndarray) -> np.ndarray:
    return (np.log2(prob.shape[1]) - entropy(prob.transpose((1, 0, 2, 3)), base=2)) / np.log2(prob.shape[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/6 19:33
@File    : cfg2tune.py
@Software: PyCharm
@Desc    : 
"""
import os
import os.path as osp
import pickle
from typing import Iterable, Callable, Any, Generator, cast, Union

from .utils import name_param_val, norm_param_name
from ..config import Config, auto_rslt_dir, ItemLazy, open_config

__all__ = ["Param2Tune", "ParamLazy", "PL", "Cfg2Tune", "P_DEP"]

kExplicitCapsKey = 'cfgs_update_at_parser'


class Param2Tune(object):

    def __init__(self, optional_values: Iterable, subject_to: Callable[[Any], bool]=lambda x: True,
                 optional_value_names: Iterable[str]=None, priority: int=None):
        """A Parameter need to be tuned

        Args:
            optional_values: Parameters optional value.
            subject_to: Filter used to check whether the parameter value is legal.
            optional_value_names: Name of optional values.
        """
        self.optional_val = tuple(optional_values)
        if optional_value_names is not None:
            self.optional_val_name = tuple(norm_param_name(name) for name in optional_value_names)
            assert len(self.optional_val) == len(self.optional_val_name)
        else:
            self.optional_val_name = tuple(name_param_val(val) for val in self.optional_val)
        self.subject_to = subject_to

        self._cur_val = None  # Current Param Value
        self._cur_val_name = None

        self._priority = priority

    @property
    def is_priority_set(self) -> bool:
        return self._priority is not None

    @property
    def priority(self) -> int:
        assert self._priority is not None
        return self._priority

    @priority.setter
    def priority(self, priority: int) -> None:
        assert self._priority is None
        self._priority = priority

    @property
    def cur_val(self) -> Any:
        if self._cur_val is None:
            raise RuntimeError(f"The current value of param {self} is None. Check if the former param call later param"
                               f"in it's subject_to.")
        return self._cur_val

    @property
    def cur_val_name(self) -> str:
        if self._cur_val_name is None:
            raise RuntimeError(f"The current value name of param {self} is None. Check if the former param call later "
                               f"param in it's subject_to.")
        return self._cur_val_name

    def reset(self) -> None:
        self._cur_val = None
        self._cur_val_name = None

    def __iter__(self) -> Generator[tuple[Any, str], None, None]:
        for opt_val, opt_val_name in zip(self.optional_val, self.optional_val_name):
            if self.subject_to(opt_val):
                self._cur_val = opt_val
                self._cur_val_name = opt_val_name
                yield opt_val, opt_val_name
        self.reset()  # 遍历结束后自动复位。

    def __repr__(self) -> str:
        return f"{self.__class__} with priority={self._priority}: \n" \
               f"optional_val={self.optional_val}, optional_val_name={self.optional_val_name}, \n" \
               f"cur_val={self._cur_val}, cur_val_name={self._cur_val_name}"


class ParamLazy(ItemLazy):
    pass


P_DEP = PL = ParamLazy


class Cfg2Tune(Config):
    """Config to be tuned with parameters to be tuned."""

    def __init__(self, *cfgs: Union[dict, str],
                 cfgs_update_at_parser: Union[tuple[str, ...], str]=(), caps: Union[tuple[str, ...], str]=(), **kwargs):
        """支持从其他其他配置树模块路径或配置树dict初始化。所有配置树会被逐个dict_update到当前配置树上。

        Args:
            *cfgs: List[配置树所在模块|配置树]
            cfgs_update_at_parser: 解析时用于更新的基配置。
            caps: cfgs_update_at_parser的别名。
            **kwargs: 传递给Dict，不应该使用。
        """
        super(Cfg2Tune, self).__init__(*cfgs, cfgs_update_at_parser=cfgs_update_at_parser, caps=caps, **kwargs)

        for leaf in self.leaves:  # 禁止从集成中获取待调参数，可以简化逻辑、令配置简单易懂。
            if isinstance(leaf, Param2Tune):
                raise RuntimeError(f"Cfg2Tune can't init by leaf with type = f{type(leaf)}")

        assert kExplicitCapsKey not in self  # cfgs_update_at_parser若定义则必用于调参，故务必在顶层定义，不可来自导入。

    @property
    def ordered_params2tune(self) -> dict[str, Param2Tune]:
        # 按照优先级排序p2t，未被设置优先级的p2t，优先级默认为最高。
        ret = dict(sorted_params2tune := sorted(((k, l) for c, k, l in self.ckl if isinstance(l, Param2Tune)),
                                                key=lambda x: x[1].priority if x[1].is_priority_set else float('-inf')))
        if len(sorted_params2tune) != len(ret):  # 找到重复键并报错。
            keys, dup_key = set(), None
            for k, _ in sorted_params2tune:
                if k in keys:
                    dup_key = k
                    break
                keys.add(k)
            raise RuntimeError(f"key = {dup_key} for param2tune repeated.")

        return ret

    def _setitem(self, key: Any, value: Any, /) -> None:
        super()._setitem(key, value)  # frozen检查、挂载、dict.__setitem__、子树自洽、IL设置。
        value = self[key]  # 重新获取value，不能保证value在设置后不变。

        # -* 若值为待调参数，检查重复。
        if isinstance(value, Param2Tune):
            root_params2tune = self.find_root()[0].ordered_params2tune  # 若p2t的key重复，会在这里报错。

            if not value.is_priority_set:  # 待调参数只要挂上了Cfg2Tune，（若没有手动设置）必被设置待调的优先级。
                cur_max_priority_p2t = next(reversed(root_params2tune.values()))
                if cur_max_priority_p2t is value:
                    assert len(root_params2tune) == 1
                    value.priority = 0
                else:
                    value.priority = cur_max_priority_p2t.priority + 1

    @staticmethod
    def dfs_params2tune(params2tune: list[Param2Tune], is_root: bool=True):
        """协程每找到一个新的参数组合，yield一次。"""
        assert len(params2tune) > 0, "The params2tune must have more than 1 Param2Tune."

        if is_root:  # 如果为外部调用，先手动复位一次。
            for param in params2tune:
                param.reset()

        cur_param = params2tune[0]

        for _ in cur_param:
            # 若只剩下一个参数，则每找到一个新参数值，就yield。
            if len(params2tune) == 1:
                yield
            # 否则遍历参数组中，第一个参数的所有可选值。对剩余参数，每找到一个新参数组合，就yield一次。
            else:
                yield from Cfg2Tune.dfs_params2tune(params2tune[1:], is_root=False)

    @property
    def param_combs(self) -> list[dict[str, tuple[Any, str]]]:
        params2tune = self.ordered_params2tune
        assert len(params2tune) > 0, "The Cfg2Tune must have more than 1 Param2Tune."

        # * 得到所有参数组合[{param1: (val1, val1_name), ...}, {param1: (val2, val2_name), ...}, ...]
        param_combs = []
        for _ in self.dfs_params2tune(list(params2tune.values()), is_root=True):
            param_combs.append({k: (v.cur_val, v.cur_val_name) for k, v in params2tune.items()})
        assert len(param_combs) > 0, "The param_combs must have more than 1 legal param comb."

        return param_combs

    @property
    def cfgs_update_at_parser(self):
        """
        Cfg2Tune支持在顶层（参见__init__，不可来自导入），以k-v对形式指定_cfgs_update_at_parser。如此可支持对
        _cfgs_update_at_parser的调参。

        当cfgs_update_at_parser来自外置k-v对时，则不应该在初始化时，不应再次指定_cfgs_update_at_parser。

        Cfg2Tune本身不会parse，所以其_cfgs_update_at_parser只用于传递给子配置。注意子配置不支持k-v对形式的
        _cfgs_update_at_parser。

        Config不会支持k-v对形式的_cfgs_update_at_parser。更新时，k-v对形式的_cfgs_update_at_parser会被覆盖（尽管不影响
        最终结果，但会造成混乱）。
        """
        if kExplicitCapsKey in self:
            assert self.get_attribute('_cfgs_update_at_parser') == ()
            if isinstance(cfgs_update_at_parser := self[kExplicitCapsKey], Param2Tune):
                return cfgs_update_at_parser.cur_val
            else:
                return cfgs_update_at_parser
        else:
            return self.get_attribute('_cfgs_update_at_parser')

    def _cfg_tuned(self) -> Config:
        """根据当前待调参数的值，返回一个Addict配置。"""
        other = Config()

        self.copy_cfg_attr_to(other)  # 将依赖、是否不可分等属性拷贝给other。
        # cfgs_update_at_parser可能来自字段，需要单独解析。
        other.set_attribute('_cfgs_update_at_parser', self.cfgs_update_at_parser)

        for k, v in self.items():
            if k == kExplicitCapsKey:
                continue  # k-v对形式的_cfgs_update_at_parser总是以attr形式传递给子配置。

            if self.is_subtree(v, self):
                other[k] = v._cfg_tuned()
            elif isinstance(v, Param2Tune):
                cur_val = v.cur_val
                if isinstance(cur_val, Config):
                    cur_val = cur_val.branch_copy()
                other[k] = cur_val
            else:
                other[k] = v

        return other

    def cfg_tuned(self, rslt_dir_suffix: str) -> Config:
        """根据当前待调参数的值，返回一个Addict配置。"""
        ret = self._cfg_tuned()
        ret.rslt_dir = osp.join(ret.rslt_dir, rslt_dir_suffix)
        return ParamLazy.compute_item_lazy(ret)

    def get_cfgs(self) -> Generator[Config, None, None]:
        """遍历待调参数的所有可能组合，对每个组合，返回其对应的配置。"""
        assert 'rslt_dir' in self

        params2tune = self.ordered_params2tune

        if len(params2tune) == 0:
            raise RuntimeError(f"The {self} must have more than 1 Param2Tune")

        for _ in self.dfs_params2tune(list(params2tune.values()), is_root=True):
            rslt_dir_suffix = ",".join([f"{name}={param.cur_val_name}" for name, param in params2tune.items()])
            yield self.cfg_tuned(rslt_dir_suffix)

    def dump_cfgs(self, cfg_dir='configs') -> tuple[list[str], list[Config]]:
        """遍历待调参数的所有组合，将每个组合对应的配置，以json（如果可以）和pkl格式保存到cfg_dir下。"""
        cfg_files, cfgs = [], []
        for cfg in self.get_cfgs():
            cfg_save_dir = osp.join(cfg_dir, cfg.rslt_dir)
            os.makedirs(cfg_save_dir, exist_ok=True)

            cfg.save_py(osp.join(cfg_save_dir, 'cfg.log'))

            cfg_file = osp.join(cfg_save_dir, 'cfg.pkl')
            cfg.save_pkl(cfg_file, save_copy=False)

            cfg_files.append(cfg_file)
            cfgs.append(cfg)

        assert len(cfg_files) > 0, "No cfg files saved."

        return cfg_files, cfgs

    @staticmethod
    def load_cfg2tune(cfg2tune_py: str, experiments_root: str='', config_root: str='./configs') -> 'Cfg2Tune':
        """导入、运行含有config = Cfg2Tune()的py文件，返回该config。"""
        '''
        若config的值，或者其Param2Tune的可选值中含有函数（也就是说cfg_tuned值含有函数），如何确保函数被正确序列化与反序列化？
        - 若函数为外部导入，譬如“from *** import func”，则确保load时，能正确执行“from *** import func”即可。
        - 若函数在config中定义，则保证：
            1）cfg2tune_py按照下面cfg2tune_import_path被import。
            2）load是也能import cfg2tune_import_path。
        注意，subject_to函数、ParamLazy函数不会被pickle，二者只在生成Addict时被执行。
        '''
        cfg2tune, _ = cast(tuple[Cfg2Tune, bool], open_config(cfg2tune_py))

        if cfg2tune.get('rslt_dir', ...) is ...:
            cfg2tune['rslt_dir'] = auto_rslt_dir(cfg2tune_py,
                                                 experiments_root=experiments_root, config_root=config_root)

        return cfg2tune



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/2/4 22:11
@File    : swa_schedulers.py
@Software: PyCharm
@Desc    : 
"""
from typing import Callable

import warnings
import math

from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer

__all__ = ['SWACycleLR', 'SWAAnnealLR']


def _linear_trans(t: float) -> float:  # 0-1等射到0-1。
    return t


def _cos_trans(t: float) -> float:  # 0-1cos变换到0-1。
    return (1 - math.cos(math.pi * t)) / 2


def _lrs_mul_factor(lrs: list[float], lr_factor: float | list[float]) -> list[float]:
    if isinstance(lr_factor, float):
        lr_factor = [lr_factor] * len(lrs)
    return [l * f for l, f in zip(lrs, lr_factor, strict=True)]


def _descend_lr_by_t(lrs_start: list[float], lrs_end: list[float], t: float) -> list[float]:
    return [lr_start * (1 - t) + lr_end * t for lr_start, lr_end in zip(lrs_start, lrs_end, strict=True)]


class SWACycleLR(_LRScheduler):
    def __init__(self, optimizer: Optimizer,
                 period_iters: int,
                 lr_start_factor: float | list[float]= 1., lr_end_factor: float | list[float]=0.,
                 start_factor_rel_base: bool=False, end_factor_rel_base: bool=False,
                 anneal_func: Callable[[float], float] | str='linear',
                 last_epoch: int=-1, **_):
        assert period_iters >= 2
        self.period_iters = period_iters

        self.lr_start_factor = lr_start_factor
        self.lr_end_factor = lr_end_factor
        self.start_factor_rel_base = start_factor_rel_base
        self.end_factor_rel_base = end_factor_rel_base

        self.lrs_start: list[float] | None = None
        self.lrs_end: list[float] | None = None

        match anneal_func:
            case 'linear':
                self.anneal_func = _linear_trans
            case 'cos':
                self.anneal_func = _cos_trans
            case _:
                assert callable(anneal_func)
                self.anneal_func = anneal_func

        super(SWACycleLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        cur_lrs = [group["lr"] for group in self.optimizer.param_groups]
        if self.last_epoch == 0:  # * 初始化时不动如山。
            return cur_lrs
        if self.last_epoch == 1:  # * 接手时确定学习率起始点。
            self.lrs_start = _lrs_mul_factor(self.base_lrs if self.start_factor_rel_base else cur_lrs,
                                             self.lr_start_factor)
            self.lrs_end = _lrs_mul_factor(self.base_lrs if self.end_factor_rel_base else self.lrs_start,
                                           self.lr_end_factor)
        t = self.anneal_func(((self.last_epoch - 1) % self.period_iters) / self.period_iters)
        return _descend_lr_by_t(self.lrs_start, self.lrs_end, t)

    def __repr__(self):
        return f"SWACycleLR <{self.__class__}>: \n" \
               f"   period_iters: {self.period_iters}\n" \
               f"   lr_start_factor: {self.lr_start_factor}\n" \
               f"   lr_end_factor: {self.lr_end_factor}\n" \
               f"   lrs_start: {self.lrs_start}\n" \
               f"   lrs_end: {self.lrs_end}"


class SWAAnnealLR(_LRScheduler):
    def __init__(self, optimizer: Optimizer,
                 anneal_iters: int,
                 lr_start_factor: float | list[float] = 1., lr_end_factor: float | list[float] = 0.,
                 start_factor_rel_base: bool=False, end_factor_rel_base: bool=False,
                 anneal_func: Callable[[float], float] | str = 'linear',
                 last_epoch: int = -1, **_):
        assert anneal_iters >= 1
        self.anneal_iters = anneal_iters

        self.lr_start_factor = lr_start_factor
        self.lr_end_factor = lr_end_factor
        self.start_factor_rel_base = start_factor_rel_base
        self.end_factor_rel_base = end_factor_rel_base

        self.lrs_start: list[float] | None = None
        self.lrs_end: list[float] | None = None

        match anneal_func:
            case 'linear':
                self.anneal_func = _linear_trans
            case 'cos':
                self.anneal_func = _cos_trans
            case _:
                assert callable(anneal_func)
                self.anneal_func = anneal_func

        super(SWAAnnealLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        cur_lrs = [group["lr"] for group in self.optimizer.param_groups]
        if self.last_epoch == 0:  # * 初始化时不动如山。
            return cur_lrs
        if self.last_epoch == 1:  # * 接手时确定学习率起始点。
            self.lrs_start = _lrs_mul_factor(self.base_lrs if self.start_factor_rel_base else cur_lrs,
                                             self.lr_start_factor)
            self.lrs_end = _lrs_mul_factor(self.base_lrs if self.end_factor_rel_base else self.lrs_start,
                                           self.lr_end_factor)
        t = self.anneal_func(min((self.last_epoch - 1) / self.anneal_iters, 1.))
        return _descend_lr_by_t(self.lrs_start, self.lrs_end, t)

    def __repr__(self):
        return f"SWAAnnealLR <{self.__class__}>: \n" \
               f"   period_iters: {self.anneal_iters}\n" \
               f"   lr_start_factor: {self.lr_start_factor}\n" \
               f"   lr_end_factor: {self.lr_end_factor}\n" \
               f"   lrs_start: {self.lrs_start}\n" \
               f"   lrs_end: {self.lrs_end}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: prefetcher.py
@time: 2020/3/20 21:22
@desc: See also https://github.com/NVIDIA/apex/issues/304
"""
from typing import Iterator
from collections import abc
import warnings
import torch

from alchemy_cat.py_tools import indent


class Prefetcher(object):
    """Wrapper for iter(data_loader).

    Return an prefetcher iter which won't block the current stream when load data.
    """

    def __init__(self, iterator: Iterator):
        """Return an prefetcher iter which won't block the current stream when load data.

        Args:
            iterator: iter(data_loader)
        """
        if not torch.cuda.is_available():
            raise RuntimeError(f"Prefetcher can only work when cuda is available")

        if not isinstance(iterator, abc.Iterator):
            raise ValueError(f"iter f{iterator} is supposed to be an iterator")

        self.iter = iterator
        self.feed_stream = torch.cuda.Stream()
        self.batch = None

        self.preload()

    def preload(self):
        try:
            self.batch = next(self.iter)
        except StopIteration:
            self.batch = None
            return
        with torch.cuda.stream(self.feed_stream):
            def prefetch(batch):
                if isinstance(batch, torch.Tensor):
                        return batch.cuda(non_blocking=True)
                elif isinstance(batch, list):
                    return [prefetch(elem) for elem in batch]
                else:
                    return batch

            self.batch = prefetch(self.batch)

    def __next__(self):
        torch.cuda.current_stream().wait_stream(self.feed_stream)
        batch = self.batch

        if batch is None:
            raise StopIteration

        def record(batch):
            if isinstance(batch, torch.Tensor):
                # Make sure tensors created by current stream won't occupied the memory of tensors from other steam
                batch.record_stream(torch.cuda.current_stream())
            elif isinstance(batch, list):
                for elem in batch:
                    record(elem)
            else:
                pass

        record(batch)

        self.preload()
        return batch

    def __iter__(self):
        warnings.warn(f"Prefetcher is already an iterator.")
        return self

    def __repr__(self):
        return f"Prefetcher: {self.__class__.__name__}\n" \
                + indent(f"iter: {self.iter}")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/7/15 17:14
@File    : read_mmcv_metric.py
@Software: PyCharm
@Desc    : 
"""
import json
import os
from pathlib import Path

__all__ = ['get_metric']


def get_metric(rslt_dir: str | os.PathLike) -> dict[str, float] | None:
    scalar_jsons = list(Path(rslt_dir).glob('**/scalars.json'))
    if len(scalar_jsons) == 1:
        final_scalar = json.loads(tuple(scalar_jsons[0].open())[-1])
        return {k: final_scalar[k] for k in ('aAcc', 'mIoU', 'mAcc')}
    elif len(scalar_jsons) == 0:
        return None
    else:
        raise RuntimeError(f'Found more than one `scalars.json` in {rslt_dir}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/7/15 15:02
@File    : mm_cfg2tune.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat import Cfg2Tune, Param2Tune, DEP, P_DEP

# Inherit from standard mmcv config.
cfg = Cfg2Tune(caps='py_tools/config/tests/migrate/mm_configs/'
                    'deeplabv3plus/deeplabv3plus_r50-d8_4xb4-40k_voc12aug-512x512.py')

# Inherit and override
cfg.model.auxiliary_head.loss_decode.loss_weight = 0.2

# Tuning parameters: grid search max_iters and batch_size
cfg.train_cfg.max_iters = Param2Tune([20_000, 40_000])
cfg.train_dataloader.batch_size = Param2Tune([8, 16])

# Dependencies:
# 1) end of param_scheduler increase with max_iters
# 2) learning rate increase with batch_size
cfg.param_scheduler = P_DEP(lambda c: [
    dict(
        type='PolyLR',
        eta_min=1e-4,
        power=0.9,
        begin=0,
        end=c.train_cfg.max_iters,
        by_epoch=False)
])
cfg.optim_wrapper.optimizer.lr = DEP(lambda c: (c.train_dataloader.batch_size / 8) * 0.01)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py.py
@time: 2020/1/7 17:12
@desc:
"""
from alchemy_cat.data.plugins.augers import *
from alchemy_cat.data.plugins.file_id_dt import *
from alchemy_cat.data.plugins.sub_augers import *
from .few_shot_dt import *
from .typing import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py.py
@time: 2020/2/25 11:56
@desc:
"""
from .warnup_lr import *
from .polynomial_lr import *
from .swa_schedulers import *
from .scheduler_chainer import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/25 20:50
@File    : cfg.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

from alchemy_cat.py_tools import Cfg2Tune, Param2Tune


def func_a(goo):
    return goo


def func_b(goo):
    return func_a(goo)


def func_c(goo):
    np.array(goo)


def func_d():
    return func_d


cfg = config = Cfg2Tune()

cfg.rslt_dir = 'def_inside_config'

# * 文件内定义的函数。
cfg.foo0.foo1.a = Param2Tune([func_a, func_b, func_c, func_d])
# * Param on diff Cfg2Tune
cfg.foo0.fix = Param2Tune(['foo.bar'])
# * Param on root Cfg2Tune
cfg.c = Param2Tune([False, True])
# * 静态值，等于文件内定义的函数。
cfg.foo0.a_max = func_a
cfg.foo0.a_min = (func_c, func_d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/2/4 16:56
@File    : warnup_lr.py
@Software: PyCharm
@Desc    : 
"""
import warnings

from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer

__all__ = ['WarmUpLR']


class WarmUpLR(_LRScheduler):

    def __init__(self, optimizer: Optimizer, last_epoch: int=-1, warm_up_iter: int=0, warm_up_min_lr: float=.0):
        self.warm_up_iter = warm_up_iter
        self.warm_up_min_lr = warm_up_min_lr
        super().__init__(optimizer, last_epoch)

    def _warm_up_ascend(self, lr: float):
        return lr * ((self.last_epoch + 1) / (self.warm_up_iter + 1))

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        assert -1 <= self.last_epoch <= self.warm_up_iter  # 该区间能给出合法学习率。
        return [max(self._warm_up_ascend(lr), self.warm_up_min_lr) for lr in self.base_lrs]
